#include <getopt.h>
#include "reltunnel.h"

#define LISTEN	1
#define CONNECT 2



struct socket_struct	sockets[MAX_SOCKETS] ;
struct packetlog 	sendlog[MAX_PACKETS_INQ] ;
struct packetlog 	recvlog[MAX_PACKETS_INQ] ;

/* extern u_int32_t curr_SEQ = SEQ_BEGIN ; */



static void print_usage (char *me) ;
static void print_version () ; 


int main (int argc, char *argv[])
{
   int c = 0 ;
   int tempi = 0 ;
   char *foo ;
   char *feh ;
   
   
   foo = (char *) malloc(256) ;
   conf.tunnel_host = (char *) malloc(256) ;
   conf.peer_host = (char *) malloc(256) ;

   
   while (1) {
      int option_index = 0 ;
      
      static struct option long_options[] =
	{
	     {"tunnel", required_argument, 0, 'T'}, 
             {"peer", required_argument, 0, 'P'},
	     {"id", required_argument, 0, 'I'},
	     {"rid", required_argument, 0, 'i'},
	     {"help", no_argument, 0, 'h'},
             {"version", no_argument, 0, 'V'},
             {0, 0, 0, 0}
	} ;
      
      
      c = getopt_long(argc, argv, "T:P:I:i:hV",
		      long_options, &option_index) ;
      
      if (c == EOF)
	break ;
      
      switch(c) {
       case 'T':	/* --tunnel | -T [tcp:hostname:port|unix:hostname:socket|..] */
	 sprintf(foo, "%s", optarg) ;
	 feh = strsep(&foo, ":") ;
	 if (!feh) {
	    fprintf(stderr, "main:\t\t\t%s %s : argument format error\n", argv[0], optarg) ;
	    exit(1) ;
	 }
	 
	    if (index(foo, ':')) { /* host and port/socket specified */
	       conf.tunnel_host = strsep(&foo, ":") ;
	       conf.tunnel_port = atoi(foo) ;
	       conf.tunnel_mode = CONNECT ; 
	    } else {      	  /* just port */
	       conf.tunnel_port = atoi(foo) ;
	       conf.tunnel_mode = LISTEN;
	    }


	 if (strcmp(feh, "tcp")==0)
	   conf.tunnel_prot = get_proto_number("tcp") ;
	 else if (strcmp(feh, "stdin")==0)
	   conf.tunnel_prot = 99 ;
	 else {
	    fprintf(stderr, "main:\t\t\t%s %s : Unsupported protocol: %s\n", argv[0], optarg, feh) ;
	    exit(1) ;
	 }
	 
	    
	   
		
		
	   
	 fprintf(stderr, "main:\t\t\ttunnel_port=%d and tunnel_host='%s' protocol=%d\n", 
		 conf.tunnel_port, conf.tunnel_host, conf.tunnel_prot) ;
	 break ;
	 
       case 'P':	/* --peer | -P [tcp:hostname:port|unix:hostname:socket|..] */
         sprintf(foo, "%s", optarg) ;
	 feh = strsep(&foo, ":") ;
	 if (!feh) {
	    fprintf(stderr, "main:\t\t\t%s %s : argument format error\n", argv[0], optarg) ;
	    exit(1) ;
	 }

	 if (index(foo, ':')) { /* host and port/socket specified */
	    conf.peer_host = strsep(&foo, ":") ;
	    conf.peer_port = atoi(foo) ;
	    conf.peer_mode = CONNECT ;
	 } else {              /* just port */
	    conf.peer_port = atoi(foo) ;
	    conf.peer_mode = LISTEN;
	 }
	 
         if (strcmp(feh, "tcp")==0)
	   conf.peer_prot = get_proto_number("tcp") ;
	 else if (strcmp(feh, "stdin")==0)
	   conf.peer_prot = 99 ;
	 else {
	    fprintf(stderr, "main:\t\t\t%s %s : Unsupported protocol: %s\n", argv[0], optarg, feh) ;
	    exit(1) ;
	 }
	 
         fprintf(stderr, "main:\t\t\tpeer_port=%d and peer_host='%s' protocol=%d\n", 
		 conf.peer_port, conf.peer_host, conf.peer_prot); 
	 break ;
			  
       case 'I':	/* --id | -I num */
	 tempi = atoi(optarg) ;
	 fprintf(stderr, "main:\t\t\tid=%d\n", tempi) ;
	 break ;
	 
       case 'i':   	/* --rid | -i num */
	 tempi = atoi(optarg) ;
	 fprintf(stderr, "main:\t\t\tremote id=%d\n", tempi) ;
	 break ; 
	 
       case 'h':	/* --help | -h */
	 print_usage(argv[0]) ;
	 exit(0) ;

	 
       case 'V':	/* --version | -V */
	 print_version();
	 exit(0) ;
	 
       default:
	 fprintf(stderr, "main:\t\t\tFOOBAR!\n") ;
	 exit(1) ;
	 
      } /* switch(c) */
	
   } /* while(1) */
   
   
   if (sizeof(struct relhdr) != RELHDR_LEN)
     die("struct relhdr size does not match protocol specs. "
         "(Did you really compile this with gcc?)") ;
   
   

   


   
   
   if (1) {
      int sp, nsp, st, nst ;

      sp = create_socket(SOCK_STREAM) ; /* peer */
      st = create_socket(SOCK_STREAM) ; /* tunnel */

      /* check peer mode of operation. 
       * (either listen and wait for connection or connect to peer )
       */
      switch(conf.peer_mode) {
       case LISTEN:
	 bind_port(sp, conf.peer_port) ;
	 start_listening(sp) ;
	 break ;
	 
       case CONNECT:
	 forward_connect(sp, conf.peer_port, conf.peer_host) ;
	 break ;
	 
       default:		/* unknown mode. bail out */
	 fprintf(stderr, "main:\t\t\tUnknown peer-mode: %i\n",
		 conf.peer_mode);
	 exit(1) ;
      }
      
	
      /* and now check tunnel operation mode. */
      switch(conf.tunnel_mode) {
       case LISTEN:
	 bind_port(st, conf.tunnel_port) ;
	 start_listening(st) ;
	 break ;
	 
       case CONNECT:
	 forward_connect(st, conf.tunnel_port, conf.tunnel_host) ;
	 break ;
	 
       default:		/* unknown mode. bail out */
	 fprintf(stderr, "main:\t\t\tUnknown tunnel-mode: %i\n",
		 conf.tunnel_mode) ;
	 exit(1) ; 
      }
      
      
      fprintf(stderr, "main:\t\t\tBoth sockets (sp, st) set up.\n") ;

      
      /* if we're in LISTEN mode on tunnel socket we'll accept 
       * connections on it before the peer socket
       * as this side probably connects first */
      if (conf.tunnel_mode == LISTEN)
	nst = await_connection(st) ;
      else
	nst = st ;
      
      if (conf.peer_mode == LISTEN)
	nsp = await_connection(sp) ;
      else 
	nsp = sp ; 
	
	
      /* both sides connected. lets start forwarding */
	
      while (1) {
	 forward_socket_data(nsp, nst) ;
      }

      
   }
	   
   
   
   
   
       

   return 0 ;
}


void print_usage(char *me) 
{
   fprintf(stderr, "usage: %s --tunnel OPTS --peer OPTS -I MYID -i OTHID\n\n", me) ;
}

void print_version() 
{
   fprintf(stderr, "reltunnel v%d.%d.%d - offers a reliable data path over unreliable protocols\n", VERSION_MAJOR, VERSION_MINOR, VERSION_REV) ;
}

